#!/usr/bin/env python3

from fire import Fire
import math

import numpy as np
import matplotlib.pyplot as plt


# https://en.wikipedia.org/wiki/Erlang_distribution#/media/File:Erlang_dist_cdf.svg
#num = input("Enter a number: ")
#print("The factorial of ", num, " is : ")
#print(math.factorial(int(num)))



# considerations
"""
 RUN 0057   1st minute -----------------------
 -  fit erlang exponential:
    1.07839e+01   ; slope    -8.96373e-02 +- 4.88916e-05
    TF1 *fa1 = new TF1("fa1","exp(1.07845e+01 -x*8.96373e-02)",0,40);

  from erlang:
   8.96946e+04 = 89 694 +- 49
   real is 85292, but if i multiply by unseen: 1.036*  => 88400 ... pretty close
   theory 51.9%DT;  caen gives 49.5%; 53.08 from adding


   RUN57 last minute ------------------

   observed rate 21546.1   theoretical from fit 21727.8 +- 38

 FCN=1199.32 FROM MIGRAD    STATUS=CONVERGED      47 CALLS          48 TOTAL
                     EDM=1.98023e-09    STRATEGY= 1      ERROR MATRIX ACCURATE
  EXT PARAMETER                                   STEP         FIRST
  NO.   NAME      VALUE            ERROR          SIZE      DERIVATIVE
   1  Constant     7.94577e+00   1.63577e-03   1.62246e-05   6.53063e-02
   2  Slope       -2.17278e-02   3.84162e-05   3.81054e-07   2.09419e+00

   the observed rate 21546 compensated on 0.986% of erlangblind = 21758.44356 - 21728 = 30...well in 1 sigma
  16.18% caen;   17.169 % from sum of 3


  a/  it seems i know rates, HOWEVER, here is a problem with double zeroes..maybe the peak
  b/  but if i do 1.5 - 4.48us window,
    - erlang says 6.021%,
    - root 6.16%
    - root with erlang 1.009 compensation ... 6.11%  PERFECT!
  THE PEAK....

"""



# probability density function   k==1 for my case
def erlang_pdf(x,rate,k=1):

    # lambda ^k    x^k-1   exp(-lambda x)
    res = np.power(rate,k )* np.power(x, k-1 )* np.exp(-rate*x)
    res = res/ math.factorial( int(k-1) )
    return res





# cumulative distribution function
def erlang_cdf(x,rate,k=1):
    return erlang_cum(x,rate,k=1)

def erlang_cum(x,rate,k=1):
    # in the test - i supply:   1/mu == rate
    res = 0
    for n in range(k):
        res+=1/math.factorial(int(n)) * np.exp(-rate*x)*np.power(x*rate,n)
        #print(f"                 n={n},  {res}")
        #res=math.exp(-rate*x)
    return 1 - res




#======================================== main ====== test
def main():
    print("Erlang_cum for rate 85000 at 8us:", erlang_cum(8e-6, 85000, 1) ," " )
    print("Erlang_cum for rate 85000 at 4.5us:", erlang_cum(4.5e-6, 85000, 1) ," " )
    print("Erlang_cum for rate 85000 at 0.4us:", erlang_cum(0.4e-6, 85000, 1) ," " )


    print("Erlang test image")
    x = np.linspace(0,20,50)

    y = erlang_cum( x, 1/2, 1 )    # mu in image == lambda here; k==k
    plt.plot(x,y,'r-')

    y = erlang_cum( x, 1/1, 9 )
    plt.plot(x,y,'b-')

    y = erlang_cum( x, 1/1, 1 )
    plt.plot(x,y,'m-')

    y = erlang_cum( x, 1/0.5, 7 )
    plt.plot(x,y,'k-')

    y = erlang_cum( x, 1/2, 3 )
    plt.plot(x,y,'y-')

    y = erlang_cum( x, 1/1, 5 )
    plt.plot(x,y,'g-')

    y = erlang_cum( x, 1/2, 2 )
    plt.plot(x,y,'-', color='orange')






    y = erlang_pdf( x, 1/2, 1 )    # mu in image == lambda here; k==k
    plt.plot(x,y,'r:')

    y = erlang_pdf( x, 1/1, 1 )
    plt.plot(x,y,'m:')

    y = erlang_pdf( x, 1/0.5, 7 )
    plt.plot(x,y,'k:')

    y = erlang_pdf( x, 1/2, 2 )
    plt.plot(x,y,':', color='orange')

    plt.margins(0.005, tight=True)
    plt.grid(True)
    plt.show()


    # for x in np.arange(0, 1.2,  0.2):
    #     print(f"{x:.1f}  (rate={rate})", erlang(x,rate))

if __name__=="__main__":
    Fire(main)



#!/usr/bin/env python3
"""
load and save operations
* load a textfile with LT A and B given in the filename
* verify under/overfl bin present
____
* definition of .details file INSIDE save_hist_txt
* real usage of .details file in     load_hist_txt with load_details
 : these are used in load_hist_txt
 - livetime
 - a
 - b
 - entries
 - rate
"""
from fire import Fire

import glob

from gregory_online.histo_global_cont import histograms

from gregory_online.histo_analyze import get_calib_coefs


from console import fg, bg, fx
import os
import numpy as np
import re # find livetime and a and b


#from gregory_online import histo_np_ops
from gregory_online.histo_np_ops import get_np_from_h
from gregory_online.histo_np_ops import fill_h_with_np
from gregory_online.histo_np_ops import get_range_from_calib

import datetime as dt

import pandas as pd

# --------------------------------------- for  tests -----------
import ROOT #  for tests here with gDirectory ls
import time
# JUST FOR TESTS
from  gregory_online.histo_displ import display_np, refresh_canvases
from gregory_online.histo_analyze import histo_fit,histo_zoomenergy
# --------------------------------------- for  tests -----------

import gregory_online.ascgeneric as ascgeneric

import configparser


# import ascgeneric
import requests
from bs4 import BeautifulSoup

def ispure2n( slen ,silent = False):
    """
    detect if it is 2^n bins or 2^n + 2 under/ovelrflow added
    """
    expon = 0
    pure2n = True
    for i in range(18): # max 2^18 ADC
        lend2 = slen/2**i
        if lend2<1.0:
            expon-=1
            break
        isint = lend2==int(lend2)
        if not isint:  pure2n = False
        # print("D... div2", expon,len(spe), lend2 , isint )
        expon+=1
    if not silent:
        print(f"D... ... ... : 2**{expon}=={2**expon} len=={slen} len==2^n: {pure2n}")
    return pure2n



def save_hist_txt( name , total_time , started, deadtime , errors = False):
    """
    if errors false:  one column, bin 1 to the last bin
    if errors true: three columns i,cont, err
    *.details file ... with DT LT ...
    """
    stamp = dt.datetime.now().strftime("%Y%m%d_%H%M%S")
    fname = f"{name}_{stamp}"
    bmin =  1 # BIN 0 is underflow
    bmax =  histograms[ name ].GetXaxis().GetNbins() # last bin with data
    with open( fname+".txt", "w" ) as f:
        for i in range( bmin, bmax+1): # the last is not included..
            cont = histograms[ name ].GetBinContent( i )
            err={}
            if errors:
                err = histograms[ name ].GetBinError( i )
                f.write(f"{i} {cont} {err}\n")
            else:
                f.write(f"{cont}\n")
    print("i... 1/2 SAVED", fname+".txt")
    cont1 = histograms[ name ].GetBinContent( 0 )
    cont2 = histograms[ name ].GetBinContent( bmax+1 )
    print(f"D...  underfl={cont1} overflow={cont2}")
    with open( fname+".details", "w" ) as f:
        TTS = total_time
        DTP = deadtime*100
        DTS = total_time*deadtime
        LTS = TTS - DTS
        SUM = histograms[ name ].GetEntries()
        RATE = SUM/TTS
        now = dt.datetime.now().strftime("%Y-%m-%d %H:%M:%S")
        if RATE>100000:
            print(f"X... {fg.red} NON REALISTIC TOTAL TIME {fg.default} ")

        a,b = get_calib_coefs( histograms[ name ])
        f.write(f"totaltime {TTS:.2f}\n")
        f.write(f"livetime {LTS:.2f}\n")
        f.write(f"deadtimeprc {DTP:.2f}\n")
        f.write(f"a {a:.6f}\n")
        f.write(f"b {b:.3f}\n")
        f.write(f"entries {SUM}\n")
        f.write(f"rate {RATE:.2f}\n")
        f.write(f"started {started}\n")
        f.write(f"saved {now}\n")
        f.write(f"name {name}\n")
        f.write(f"fname {fname}\n")
        if errors:
            f.write(f"errors_column True\n")



    print("i... 2/2 SAVED", fname+".details")

    print()
    return



# -------------------------------------- for load interface
def list_spectra_disc( folder = "./" ):
    """
    look at spectra and details files. determine valid spectrum filename
    """
    f = folder
    if len(f)==0: f = "./"
    if f[-1]!="/": f=f+"/"

    avadisc = {}
    res = None
    red = None
    if f.find("http://")==0:
        res = get_url_paths(folder, "txt")
        red = get_url_paths(folder, "details")
    else:
        res = glob.glob(f+"*.txt")
        red = glob.glob(f+"*.details")

    num = 1
    for i in res:
        tst = os.path.splitext(i)[0]
        for j in red:
            if tst == os.path.splitext(j)[0]:
                avadisc[num] = tst
                num+=1
    return avadisc


# -------------------------------------- for load interface ASC

def get_url_paths(url, ext='', params={}):
    """
    list files from NAS - http/web exposed folder
    """
    if url[-1] != "/":
        url+="/"
    response = requests.get(url, params=params)
    if response.ok:
        response_text = response.text
    else:
        return response.raise_for_status()
    soup = BeautifulSoup(response_text, 'html.parser')
    parent = [url + node.get('href') for node in soup.find_all('a') if node.get('href').endswith(ext)]
    return parent


# -------------------------------------- for load interface ASC
def list_asc_disc( folder = "./"):
    """
    look at spectra and details files. determine valid spectrum filename
    """
    f = folder
    if len(f)==0: f = "./"
    if f[-1]!="/": f=f+"/"
    avadisc = {}
    res = None
    red = None
    if f.find("http://")==0:
        res = get_url_paths(folder, "asc")
        red = get_url_paths(folder, "ini")
    else:
        res = glob.glob(f+"*.asc")
        red = glob.glob(f+"*.ini")
    num = 1
    for i in res:
        tst = os.path.splitext(i)[0]
        for j in red:
            if tst == os.path.splitext(j)[0]:
                avadisc[num] = os.path.basename(tst)
                num+=1
    return avadisc


# ================================================ ASC AND INI ==================
#
# I like it here, bc menu interaction
#
def load_asc( filename  ):  # complete filena or address
    """
    loads asc file, needs ini file too

    returns DF and inifile
    """
    # f = folder
    # if len(f)==0: f = "./"
    # if f[-1]!="/": f=f+"/"

    #avadisc = list_asc_disc( folder = folder)
    # # present_avadisc(avadisc)


    inifilename = os.path.splitext(filename)[0]+".ini"
    ascfilename = os.path.splitext(filename)[0]+".asc"

    exini = False
    if filename.find("http://")==0:
        inifilename = ascgeneric.get_ini_copy( inifilename )
        print(" ... ",inifilename)
        #exini = True

    else:
        #print("X... no file ", filename, end=" ...")
        if os.path.exists(ascfilename):
            pass
            # filename = f+hole+".asc"
            # print("i... replacing with ", filename)
        else:
            print("X... no file ", ascfilename)
            return None, None

    res = ascgeneric.load_ini_file( inifilename ) # I never need channel here.. Later, after CUT
    inifile = True

    if res is None:
        print(f"x... {bg.red}NO INI FILE PRESENT{bg.default}")
        inifile = False
        return None, None
    else:
        print(f"i... {bg.green} INI FILE PRESENT{bg.default}")
        start = ascgeneric.filename_decomp( filename )   #  - get start from filename TO REPORT
        print("D... started", start, type(start) )
        print(f" {bg.white}{fg.black} ... LOADING {ascfilename} ... {fg.default}{bg.default}")
        wastart = dt.datetime.now()
        df = ascgeneric.pd_read_table( ascfilename, sort = True,  fourcolumns = False)
        wastop = dt.datetime.now()-wastart
        print(f" {bg.white}{fg.green} ... LOADED {len(df)/1000/1000:.2f} Mrows ... {len(df)/wastop.total_seconds()/1000/1000:.2f} Mrows/sec {fg.default}{bg.default}")
        print()

        #df = pd.DataFrame( range(10) )
        #print(df)

    return df, inifilename, start


def select_asc_to_load( num=None , folder = "./"):
    """
    present available asc files and/or load one

    load_asc .... return df,filename
    """
    avadisc = list_asc_disc( folder = folder )
    if num is None:
        present_avadisc(avadisc, bgcolor = f"{bg.white}")
        return None,None,None, None

    try:
        num = int(num)
    except:
        num = None
        print("X... no spectrum selected")
    if num in avadisc.keys():
        f = folder
        if len(f)==0: f = "./"
        if f[-1]!="/": f=f+"/"

        # return DF and name
        print(f"D...  s.a.t.l. {avadisc[num]}")
        df, ininame,start = load_asc( f+avadisc[num]  )
        # returning from  select_asc_to_load - called from bin
        return  df, avadisc[num], ininame,start
    return None,None,None, None



def present_avadisc( avadisc , bgcolor=f"{bg.yellow}"):
    print()
    col = 0

    termsize = os.get_terminal_size()
    maxwid = termsize.columns
    maxname = 0
    for ke in sorted(avadisc.keys()):
        if len(avadisc[ke])>maxname:maxname = len(avadisc[ke])
    maxname+=5
    ncols=int(maxwid/maxname)
    if ncols == 0: ncols =1
    print(maxwid, maxname, ncols )
    for ke in sorted(avadisc.keys()):
        col+=1
        brk = ""
        if col% ncols==0:
            brk = "\n"
        print( f"  {bgcolor}{fg.black} {ke} {fg.default}{bg.default}  {avadisc[ke]:{maxname}s} ", end=brk)
    print()
    print()


# ================================================ HISTO AND DETAILS ==================


def select_histo_to_load( num=None , folder = "./"):
    """
    full sequence list,present, load
    """
    f = folder
    if len(f)==0: f = "./"
    if f[-1]!="/": f=f+"/"

    avadisc = list_spectra_disc( folder = folder )
    present_avadisc(avadisc)
    if num is None:
        return

    try:
        num = int(num)
    except:
        num = None
        print("X... no spectrum selected")
    if num in avadisc.keys():
        load_hist_txt( avadisc[num] , f"s{num}" )


# ----------------------------------- loading spectra/histogram

def load_details(fname):   # used in load_hist_txt: where histograms[hname] is created
                           # and rate interpretted
    res = None
    if os.path.exists(fname):
        with open(fname) as f:
            res = f.readlines()
        res = [ (x.strip().split(" ")[0],"_".join(x.strip().split(" ")[1:])) for x in res]
        res = dict(res)

        # the following are gonna be converted and cause a crash if error
        res["totaltime"] = float(res["totaltime"]) # not always needed?
        # not floatres["started"] = float(res["started"]) # not always needed
        res["livetime"] = float(res["livetime"])
        try:
            res["a"] = float(res["a"])
            res["b"] = float(res["b"])
        except:
            res["a"] = 1
            res["b"] = 0
            print(f"X... {fg.red} I dont see a calibration in the details file {fg.default}" )
        res["entries"] = float(res["entries"])
        res["rate"] = float(res["rate"])
    return res





# ----------------------------------------------------







def load_hist_txt( filename, hname , underoverflow = False):
    """
    loads histo : full or rate w. details file.
    histograms[] will get NORMALIZED version;  underover bit is appended just for TH1
    returns nparray without under over flow.
    """
    # I use pandas as in pr_load.py ...
    #avadisc = list_spectra_disc()
    #present_avadisc(avadisc)

    hfilename = os.path.splitext(filename)[0]+".txt"
    dfilename = os.path.splitext(filename)[0]+".details"
    details = True
    if filename.find("http://")==0:
        dfilename = ascgeneric.get_ini_copy( dfilename )
        res = load_details(  dfilename )
        if res is None: details = False
    else:
        res = load_details(  dfilename )
        if res is None: details = False

    # #hole = os.path.splitext(filename)[0]
    # if not os.path.exists(filename):
    #     print("X... no file ", filename, end=" ...")
    #     if os.path.exists(hole+".txt"):
    #         filename = hole+".txt"
    #         print("i... replacing with ", filename)
    #     else:
    #         print("X... no file ", filename)
    #         return None

    # res = load_details( hole+".details")
    # details = True
    # if res is None: details = False

    print("D... histo-txt:", hfilename )
    df = pd.read_csv(hfilename, delimiter=r"\s+", header=None, comment="#", nrows=2)
    ncols = df.shape[1]
    print(f"D... file {hfilename} with {ncols} columns, details found = {details}")

    np_c = None
    np_e = None # for rate into th1f

    if (ncols==1):
        df = pd.read_csv(hfilename, delimiter=r"\s+", header=None, comment="#", names = ['cont'] )
        print(df)
        np_c = df['cont'].values
    if (ncols==3):
        df = pd.read_csv(hfilename, delimiter=r"\s+", header=None, comment="#", names = ['bin','cont','err'] )
        print(df)
        np_c = df['cont'].values
        np_e = df['err'].values

        # print( type(np_c), type(np_e), np_c)
        # df.to_csv("zest.txt", sep=' ', header = None)


    spe = np_c # ================= taking content
    pure2n = ispure2n( len(spe) )
    isunovf = ispure2n( len(spe)-2, silent = True )
    if pure2n:
        print("i... ...  seems no under/overflow bins in the file")

    # ------------------------------------------ this part is tricky, FOR UND OVF LOAD
    if underoverflow: print("I... supposing the UND/OVER FLOW bins are already present!!")
    if underoverflow and isunovf:
        print("D... Under/overflow bins detected and loading")
    if not(underoverflow) and isunovf:
        print(f"X... {fg.red} Under/overflow bins DETECTED but NOT loading {fg.default}")
    #
    if underoverflow and  pure2n:
        print(f"X... {fg.red} PROBLEM - file is 2^n but loading w U/OVF {fg.default}")
    if (not underoverflow) and  (not pure2n):
        print(f"X... {fg.red} PROBLEM - file is NOT 2^n but loading wout U/OVF {fg.default}")
    # ------------------------------------------ this part was tricky, FOR UND OVF LOAD

    if not underoverflow:
        print("D... adding artificially undeflow/overflow bins...")
        np_c = np.concatenate(([0],np_c,[0]))
        if np_e is not None:
            np_e = np.concatenate(([0],np_e,[0]))
    else:
        print("D... underflow/overflow bits (must be) already present in source file")

    h1np_backg = np_c
    rate = res['rate']
    if abs(h1np_backg.sum()-rate)/rate<0.1:
        print("i... ... already RATE histogram interpretation !!! ...")
        rate = res["rate"]
        res['livetime'] = 1
    else:
        if res['livetime'] is not None and res['livetime']>0:
            rate = h1np_backg.sum()/res['livetime']
    #print(f"i... LEN={len(h1np_backg)} rate={rate}" )
    # normalize
    #print(  res["b"] )
    #print( res["a"] )
    #print( float(res["a"])*(len(h1np_backg)-2) )
    #print( res["b"]  )

    #  create a normalized histo in the histograms[]
    fill_h_with_np( h1np_backg, hname,
                                 runtime = res["livetime"], errors = np_e,
                                 limits_calib=[res["b"], res["a"]*(len(h1np_backg)-2)+res["b"] ]
    )
    histograms[hname].SetEntries( res["entries"] )
    histograms[hname].SetTitle(  os.path.basename(hfilename) )
    histograms[hname].Print()
    print(f" ... Name= {histograms[hname].GetName()}  Title={histograms[hname].GetTitle()}" )

    if not underoverflow:
        # artificialy added unovf, removing again
        return h1np_backg[1:-1]
    else:
        # already present in spe
        return h1np_backg[1:-1]




    # grp = re.search(r'[\w\d]+_lt([\d\.]+)_a([\d\.]+)_b([\-\d\.]+)\.txt',filename)
    # print(f"D... histo decoded  lt,a,b={fg.green} {grp.groups()} {fg.default}")
    # h1np_backg_lt = float(grp.groups()[0])
    # cala_bg = float(grp.groups()[1])
    # calb_bg = float(grp.groups()[2])

    # pure2n = True # is it pure 2^n?

    # if os.path.exists(filename):
    #     with open(filename) as f:
    #         spe = f.readlines()
    #     spe = [ float(x.strip()) for x in spe]
    #     # 2**15 32768

    #     #spe.append(0)
    #     #spe.append(0)
    #     pure2n = ispure2n( len(spe) )
    #     isunovf = ispure2n( len(spe)-2 )
    #     if underoverflow and isunovf:
    #         print("D... Under/overflow bins detected and loading")
    #     if not(underoverflow) and isunovf:
    #         print(f"X... {fg.red} Under/overflow bins DETECTED but NOT loading {fg.default}")
    #     # expon = 0
    #     # for i in range(18): # max 2^18 ADC
    #     #     lend2 = len(spe)/2**i
    #     #     lend2a = (len(spe)-2)/2**i
    #     #     if lend2<1.0:
    #     #         expon-=1
    #     #         break
    #     #     isint = lend2==int(lend2)
    #     #     if not isint:  pure2n = False
    #     #     # print("D... div2", expon,len(spe), lend2 , isint )
    #     #     expon+=1
    #     # print(f"D... calculated exp: 2**{expon}=={2**expon} len=={len(spe)} pure2n={pure2n}")

    #     if underoverflow and  pure2n:
    #         print(f"X... {fg.red} PROBLEM - file is 2^n but loading w U/OVF {fg.default}")
    #     if (not underoverflow) and  (not pure2n):
    #         print(f"X... {fg.red} PROBLEM - file is NOT 2^n but loading wout U/OVF {fg.default}")


    #     if not underoverflow:
    #         print("D... adding undeflow/overflow bins...")
    #         spe.insert(0,0) # to be compatible with underflow and overflow
    #         spe.append(0)
    #     else:
    #         print("D... underflow/overflow bits already present in source file")
    #     h1np_backg = np.array( spe )
    #     rate = 1
    #     if h1np_backg_lt is not None and h1np_backg_lt>0:
    #         rate = h1np_backg.sum()/h1np_backg_lt
    #     print(f"i... {h1np_backg} LEN={len(h1np_backg)} rate={rate}" )
    #     # normalize
    #     histo_np_ops.fill_h_with_np( h1np_backg, hname, h1np_backg_lt, limits_calib=[calb_bg, cala_bg*(len(h1np_backg)-2)+calb_bg ])

    # if not underoverflow:
    #     # artificialy added unovf, removing again
    #     return h1np_backg[1:-1]
    # else:
    #     # already present in spe
    #     return h1np_backg[1:-1]





def main():# fname, hname):
    """
    I can load txt file when lt a b  are in the filename...
    """
    #load_hist_txt(fname, hname)
    select_histo_to_load()
    ROOT.gDirectory.ls()

    for i in histograms.keys():
        # histograms[i].Draw()
        hnp,enp = get_np_from_h(histograms[i] )
        # this is in h already...
        a,b = get_calib_coefs( histograms[ i ])
        lo,hi = get_range_from_calib( histograms[ i ], a, b)
        print(hnp)
        print(enp)
        print("="*50)
        display_np( hnp, "q", "c_q" , errors=enp , limits_calib=(lo,hi) , filled=True, color=32)
        # display_np( hnp, "q", "c_q" , errors=enp  )
        # ------------  I can plot from here
        # ...
        # now fit test
        histo_zoomenergy( histograms[i], 1455,1468)
        #histo_zoomenergy( histograms[i], 929,939)
        res = histo_fit( histograms[i],  canvas = "fitresult")
        #print(res)
        break



    while True:  time.sleep(1)

if __name__=="__main__":
    Fire(main)



#!/usr/bin/env python3

from fire import Fire
import numpy as np
import scipy.ndimage
#from scipy.ndimage.interpolation import shift
import matplotlib.pyplot as plt



def rebin_np( a , sca, shi, order = 3):
    b = scipy.ndimage.zoom(a, sca, order=order)
    b = scipy.ndimage.interpolation.shift(b, shi , cval=0, order = order)
    asu = a.sum()
    bsu = b.sum()
    b=b/bsu*asu
    df = len(b) - len(a)
    #print( f" b={len(b)}    a={len(a)}   df = {df}")
    while df<0:
        b = np.append( b,[0] )
        df = len(b) - len(a)
    if df>0:
        b = b[:-df]
    #print( len(a),"X" ,len(b),  f" 6 -->  {6*sca+shi}" )
    #print( len(a),"X" ,len(b),  f" 22 -->  {22*sca+shi}" )
    return b




def main():

    a = np.array( [2,2,4,5,6,12,41,17,5,4,3,2,1,2,4,3,5,4,2,3,4,11,44,43,14,2,3,2,3] )
    a = a.astype(float)
    print(a)


    b = rebin_np(a, 0.9, 2.2, order=1 )
    c = rebin_np(a, 0.9, 2.2, order=3 )

    db=b-a
    dc=c-a

    plt.plot(a)
    plt.plot(b,'y--')
    plt.plot(c,'r--')
    plt.plot(db,'y:')
    plt.plot(dc,'r:')
    plt.grid()
    plt.show()
    #print(b-a)

if __name__=="__main__":
    Fire(main)



#!/usr/bin/env python3

from fire import Fire

import time

import pytermgui
from pytermgui import print_to, report_cursor, bold, inverse, underline, italic, save_cursor, restore_cursor, terminal, cursor_up, move_cursor


import threading # for key input


from console import fg, bg, fx
import datetime as dt

theight= terminal.height
twidth= terminal.width

global_mode = " "


class Topbar:
    def __init__(self, pos=1, bgcolor=bg.blue ):
        self.pos = pos
        self.positions = {}
        self.t2 = None
        if pos==1:
            self.BCOL = bgcolor #bg.blue
        elif pos==2:
            self.BCOL = bgcolor # bg.white
        #self.t = threading.currentThread()

        try:
            #print("report_cursor to appear")
            #print( "i... topbar: pos/cursor",pos, report_cursor() )
            pos
            report_cursor()
            #print("report done")
        except:
            print("X... problem with report_cursor")

        #print("i... topbar bar started")


    def add(self,two=2, bgcolor=bg.blue ):
        if two==2:
            self.t2 = Topbar(two ,bgcolor = bgcolor)
        else:
            print("X... nobody wanted more than two......  NOT OK")
        return self.t2


    def print_to(self, tup, s):
        if type(tup) is tuple:
            x,y = tup[0],tup[1]
            print("X.......... TUPLE in the TOPBAR  IS SUPRESSED")
        elif type(tup) is int:
            x=tup
            y=1
        else:
            print("X... NOBODY WANTED something else than tuple or int in the TOPBAR  position")
        self.positions[ x ] = s


    def place(self):
        '''
        Place he BAR on screen
        '''
        curs =(-1,-1)
        if self.pos==1:
            save_cursor()
        print_to( (1,self.pos),  f"{self.BCOL}"+" "*twidth+bg.default )
        print_to( (1,self.pos+1), " "*twidth )

        ###### self.positions[ twidth] = f"{fx.default}{fg.default}{bg.default}"

        for k in self.positions.keys():
            print_to( (k,self.pos), f"{self.BCOL}{self.positions[k]}{bg.default}{fx.default}{fg.default}" )



        if self.t2 is not None:
            self.t2.place()

        if self.pos==1:
            restore_cursor()
            print("", end="\r") # this make GOOD thing in printing



def main():
    #print()
    t = Topbar(1)
    for i in range(100):
        #
        # DO whatever stuff and PLACE PRINTTO SLEEP
        #
        t.place()
        t.print_to( (1,1), f"{fg.white} {str(dt.datetime.now())[:-4]} {fg.default}" )
        time.sleep(0.1)

if __name__=="__main__":
    Fire(main)



#!/usr/bin/env python3

from fire import Fire
from gregory_online.version import __version__


from pyfromroot import  prun
from console import fg,bg

import os

#
#  taken from flashcam and stackoveerflow
import math

def is_int(n):
    # try:
    #     if math.isnan(n):
    #         return False
    # except:
    #     return False

    try:
        float_n = float(n)
        int_n = int(float_n)
    except ValueError:
        return False
    else:
        return float_n == int_n


def is_float(n):
    try:
        float_n = float(n)
        #print(n, float_n)
    except ValueError:
        print(f"X... not float /{n}/ ... false ret")
        return False
    else:
        #print("true block", n)
        if n is None:
            print("X... isfloat ... NONE")
            return False

        retval = True
        try:
            if type(n) is not str and math.isnan(n): # if Nan comes, it can sort it out, else it crashes with string.
                print("X... math ... NAN float")
                retval = False
        except:
            print("X... crashed math NaN")
            retval = False
        return retval # normally true


def is_bool(n):
    if type(n) is str and n=="False": return True
    if type(n) is str and n=="True": return True
    return False




def ecalibration( calist, overridefile = ""):
    ECALIB = calist
    if len(ECALIB)==0:
        return 1,0
    if len(ECALIB)==1:
        ECALIBb = 0
        ECALIBa = ECALIB[0][0] / ECALIB[0][1]
        return ECALIBa, ECALIBb
    elif len(ECALIB)==2:
        ECALIBa = (ECALIB[0][0] - ECALIB[1][0]) / ( ECALIB[0][1] - ECALIB[1][1] )
        ECALIBb = ECALIB[0][0] - ECALIBa* ECALIB[0][1]
        return ECALIBa, ECALIBb
    else:
        print("i... LINEAR REGRESSION needed now")
        fname,fnamefull = None,None
        if overridefile == "": # ---create file to load --
            fname = "/tmp/tmp_ecalib"
            fnamefull = f"{fname}.txt"
            with open(fnamefull,"w") as f:
                for pair in ECALIB:
                    f.write( f"{pair[0]} {pair[1]} {pair[2]}\n")

        else: # OR ------** just take some other file **
            fnamefull = overridefile
            fname = os.path.splitext(fnamefull)[0]

        # and do fit
        prun.loadpy("load",f"{fnamefull} y,x")
        print(f"i...  {bg.green} _________________loaded___________{bg.default}")
        res = prun.loadpy("fit",f"{fname} pol1", canvas="ecalibration")
        #print(res)
        print()
        print(res.keys())
        return res["a"],res["b"]
        #prun.loadpy("draw","tmp_ecalib")
        #print("press enter")
        #input()
        #prun.do("draw","/tmp/tmp_ecalib", canvas = "cn2")

def main( txt , overridefile = ""):
    print( "f : ",is_float(txt) )
    print( "i : ",is_int(txt) )
    print( "b : ",is_bool(txt) )

    calist = [[1764.5, 2967.1799251154357, 0.1], [1460.8, 2456.0100724765166, 0.2], [609.3, 1023.7130926318914, 0.05], [351.9, 591.028749200901, 0.3]]
    ecalibration (calist , overridefile = overridefile)
    print("say ENTER 1")
    input()
    print("----{bg.yellow} ---------------------------- {bg.default}")
    ecalibration (calist , overridefile = overridefile)
    print("say ENTER")
    input()


if __name__=="__main__":
    Fire(main)



#!/usr/bin/env python3

from fire import Fire
#from read_vme_offline.version import __version__

#print("i... module read_vme_offline/ascdf is being run")

import pandas as pd
#import modin.pandas as pd
import tables # here, to avoid a crash when writing pandas
# import h5py not needed
import datetime
import subprocess as sp
import numpy as np

import os
import datetime as dt

import matplotlib.pyplot as plt

import ROOT

import sys
# import root_numpy# 2019 is old with newer numpy

# import psutil

import datetime as dt

from console import fg,bg,fx

import gregory_online.erlang as erlang
import gregory_online.dpp2jpg as dpp2jpg

import tempfile
import urllib.request


#
#  Works when apply is used: df.parallel_apply(func)
#
#
#from pandarallel import pandarallel
# Initialization
#pandarallel.initialize()








#------------------------------------------------------------------------

def get_number_of_lines(filename):
    """
    wc cli commad to count the lines of ASC
    """
    ps = sp.Popen( ['cat', filename], stdout=sp.PIPE)
    output = sp.check_output(('wc', '-l'), stdin=ps.stdout)
    ps.wait()
    nlines=int(output.decode("utf8").split()[0])
    return nlines

#------------------------------------------------------------------------

def get_number_of_columns(filename):
    """
    omit first 4 lines, take one line and tell
    """
    CMDL = ['head','-n','4', filename]
    ps = sp.Popen( CMDL , stdout=sp.PIPE)
    # print(" ".join(CMDL))
    output = sp.check_output(('tail','-n','1'), stdin=ps.stdout)
    ps.wait()
    ncolumns=len(output.decode("utf8").strip().split())
    if ncolumns==5:
      print(f"D... {ncolumns} cols ... 2020+ version with pile-up info (1/0)")
    elif ncolumns==4:
        print(f"D... {ncolumns} cols  ... 2018 version without extras column, pileup (3/0)")
    else:
        print(f"X... BAD NUMBER OF COLUMNS {ncolumns}")
        sys.exit()
    return ncolumns




# ----------------------------------------------ASC AND INI COPY TO TMP
def get_ini_copy(url):

    print(f"D... fetching {url}")
    filename = tempfile.NamedTemporaryFile(suffix='.ini', delete=False)
    txt = ""
    try:
        with urllib.request.urlopen( url ) as uf:
            txt = uf.read().decode('utf-8')
    except urllib.error.URLError as e:
        print("X... URL ERROR: ",e.reason)

    if len(txt)>10:
        with open( filename.name, 'w', encoding='utf8', newline='') as f:
            f.write( txt )

    return filename.name

# ==================================================== ASC AND INI =================


def load_ini_file(filename, channel = None ): # already complete filename

    inifilename = os.path.splitext(filename)[0]+".ini"
    if inifilename.find("http://")==0:
        inifilename = get_ini_copy( inifilename )


    if os.path.exists(inifilename):
        #print(f"D... {fg.green} OK - ini file exists {fg.default}")
        if channel is not None:
            print(f"D... {fg.green} OK channel {fg.default}")
            TRG_HOLDOFF,TF_PEAK_HOLDOFF, TWIN_US, ECALa, ECALb = dpp2jpg.extract_holdoffs( inifilename, channel )
            print(  TRG_HOLDOFF,TF_PEAK_HOLDOFF, TWIN_US)
            return TRG_HOLDOFF,TF_PEAK_HOLDOFF, TWIN_US, ECALa, ECALb

        return 0,0,0, 1,0 # last is calibration

    return None,None,None, 1,0 # last is calibatoin


# ==================================================== ASC AND INI =================



# ------------------- statistics for ascdf  NOT ascdf_chan

def stat_asc(df):
    print("+"*50,"BASICS TOTAL BEGIN")
    chan_available = df['ch'].unique()
    print("D... channels available:", chan_available)
    for i in chan_available:
        evts = len( df.loc[ (df['ch']==i) ])
        print(f"    {i}: {evts } events")
    print(f"D... pu        present: {df['pu'].unique()} ... 1 or 0 (3v0 before 2020)" )
    if 'extras' in df:
        print(f"D... extras    present: {df['extras'].unique()}... 1..satur,3..roll,4..reset,8..fake" )
        print(f"D... total saturations: ",len( df.loc[ (df['extras']&1)==1]  ) )
    #print( "D... Emin            :",df['E'].min() )
    #print( "D... Emax            :",df['E'].max() )
    print( "D... lasttime        :",df['time'].max() )
    print("+"*50,"BASICS TOTAL END")

    print(f" {bg.white}{fg.black} INPUT channel now{fg.default}{bg.default}")



    # print(f"D... calib           :  {CALA},  {CALB}  for chan {chan}" )

    # inifile = os.path.splitext( filename)[0]+".ini"
    # if inifile.find("http://")==0:
    #     inifile = get_ini_copy( inifile )
    # print(inifile)
    # CALA = 1
    # CALB = 0
    # if os.path.exists( inifile  ):
    #     config = configparser.ConfigParser( delimiters=(" ","\t") )
    #     config.read( inifile )
    #     CALA =config[f"{chan}"]['CALIBRATION_A']
    #     CALB =config[f"{chan}"]['CALIBRATION_B']
    #     CALA = float(CALA.split("#")[0].strip()) # avoid comments at the end of line
    #     CALB = float(CALB.split("#")[0].strip())
    #     print(f"D... INI file exists {inifile}, calibration =  {CALA} {CALB}")

    # return df






#------------------------------------------------------------------------

def filename_decomp(filename):
    """
    The ONLY procedure to get the start time from filename
    """
    # should work both with 2021mmdd and  21mmddd
    # returns the start time
    #

    basename = os.path.basename(filename)
    basename = os.path.splitext(basename)[0]
    # start-date
    startd = basename.split("_")[1]
    # start-time
    startt = basename.split("_")[2]

    # 4 digits always
    if len(startd)==6:
        print("D...  compensating 2 digit year to 4 digit year")
        startd="20"+startd

    print("D...  time start MARK=",startd+startt)


    #ok = False
    #try:
    start = dt.datetime.strptime(startd+startt,"%Y%m%d%H%M%S" )
    #ok = True
    print("D... succesfull start with 4 digit year")

    #except:
    #    print("x... year may not by 4 digits")
    #if not(ok):
    #    print("X... trying 2 digits for year")
    #    start = dt.datetime.strptime(startd+startt,"%y%m%d%H%M%S" )

    return start




#------------------------------------------------------------------------

def pd_read_table(filename, sort = True,  fourcolumns = False):
    """
    READ THE ASC TABLE:   two possibilities:  4 columns OR 5 columns
EARLIER:
         Events[ch][ev].TimeTag, ENERGY,(Events[ch][ev].Energy)>>MAXNBITS, ch
         MAXNBITS 15
Feb 2020+:
       Events[ch][ev].TimeTag, ENERGY,  (Events[ch][ev].Energy)>>MAXNBITS, ch, Events[ch][ev].Extras
       MAXNBITS 15
    >>t,e,PU (1/0) (or 3/0 old)  ,  channel,   Extras(2020+ only)


https://www.npl.washington.edu/TRIMS/sites/sand.npl.washington.edu.TRIMS/files/manuals-documentation/CAENDigitizer_SW_User_Manual_rel.5.pdf

EXTRAS: bit[0] = DEAD_TIME. This is set to 1 when a dead time occurred before this event. The dead time can be due to
either a signal saturation or a full memory status. Check Fig. B.4 and Fig. B.5 for more details
bit[1] = ROLL_OVER. Identify a trigger time stamp roll-over that occurred before this event
bit[2] = TT_RESET. Identify a trigger time stamp reset forced from external signals in S-IN (GPI for Desktop)
bit[3] = FAKE_EVENT. This is a fake event (which does not correspond to any physical event) that identifies a
time stamp roll-over. The roll-over can be due to an external or internal reset. The user can set bit[25]
= 1 of register 0x1n80 to enable the fake-event saving in case of reset from S-IN, and bit[26] = 1 of
register 0x1n80 to enable the fake-event saving in case of internal roll-over. In the first case the event
will have both bit[3] and bit[2] set to 1, while in the second case the event will have both bit[3] and
bit[1] set to 1.
    """
    #print(f"D... fourcolumns=={fourcolumns}")
    #print(f"D... fourcolumns=={fourcolumns}")
    #print(f"D... fourcolumns=={fourcolumns}")
    start = dt.datetime.now()
    df = pd.read_table( filename,
                        names=['time','E','pu','ch','extras'],
                        sep = r"\s+",
                        comment="#")

    meastime = (dt.datetime.now() - start).total_seconds()
    #print(f"D... table is in memory ...  {get_number_of_lines(filename)/meastime/1000/1000:.3f} Mrows/sec")
    print(f"D... table is in memory ...{len(df)/meastime/1000/1000:.1f} Mrows/sec")

    #print(f"D... fourcolumns=={fourcolumns}")

    #print( freemem() )
    # drop if 4 column
    df = df.dropna(axis='columns', how='all')


    df['time'] = df['time']/1e+8 # this is correct for 10ns step ADC
    # print("D... table last timestamp ",df.iloc[-1]["time"], "sec.")

    #print(df.dtypes)
    df['time'] = df['time'].astype('float64')
    df['E']    = df['E'].astype('int32')
    df['pu']    = df['pu'].astype('int32')
    df['ch']   = df['ch'].astype('int32')
    if 'extras' in df:
        df['extras'] = df['extras'].astype('int32')
    #print("D...    finale types ************************************")
    #print(df.dtypes)

    if fourcolumns:
       print("D... extra operation for four columns ...  FOURCOLUMNS TABLE !!!")
       print(df)
       df['E'] = df['E']+ (df['pu']&1) * 16384
       df['pu'] = df['pu'].apply(lambda x: x >> 1)
       print(df)
    # print(freemem())

    #    df.to_hdf('run'+str(number)+'.h5',
    #              "matrix",
    #              format='t',
    #              data_columns=True,
    #              mode='w')

    # print(freemem() )
    if sort:
        print("D... sorting (after read_table):")
        #print(df)
        df = df.sort_values(by="time")
        #print(df)
        df.reset_index(inplace=True, drop=True)
        #print(df)
        #print( freemem() )

    #print("D... table last timestamp ",df.iloc[-1]["time"], "sec.")
    #endblock()
    return df





#------------------------------------------------------------------------
def select_channels(df, chan, delete_original = False):
    """
    select channels bu not as view! COPY and DELETE
    """
    print("D... selecting channels:", chan, end="...  ", flush = True)

    channels = None
    if type(chan) is list:
        channels = chan
    else:
        channels = [chan]

    #print( freemem() )
    # CHANNEL (i hope it is sorted)
    if delete_original:
        df1 = df.loc[ df["ch"].isin( channels ) ].copy()
        del df
    else:
        df1 = df.loc[ df["ch"].isin( channels ) ]
    df1.reset_index(inplace=True, drop=True)
    #del df
    if len(df1) ==0 :
        print("X... no data for channels",channels)
        return None
        #sys.exit(1)
    return df1




def enhance_dt( df ):

    df1 = df

    print("i... enhancing DataFrame by dtus and dtus_prev ", end="...", flush=True)

    df1, dtusmin, dtusmax = enhance_by_dtus_and_next_E(df1)

    print(" ")

    tmin = df1['time'].min()
    tmax = df1['time'].max()
    interval = tmax - tmin
    chan_available = df1['ch'].unique()
    chan = chan_available[0]

    print("_"*50,"BASICS BEGIN CHANNEL ",chan)
    print("D... channels available:", chan_available)

    print(f"D... pu        present: {df1['pu'].unique()}... 1 or 0 (3v0 before 2020)" )
    #print(f"D... pu        present: {df1['pu'].unique()}... 1 or 0 (3v0 before 2020)" )
    pileups_ch = len(df1.loc[ df1['pu']==1])  # (3)for 4 columns, it is fixed in table_read
    print(f"D... pu       (pu==1) : {pileups_ch} ")
    if 'extras' in df:
        print(f"D... extras    present: {df1['extras'].unique()}... 1..satur,3..roll,4..reset,8..fake" )
        satu_n_ch = len(df1.loc[ (df1['extras']&1)!=0] )
        Tsatu_ch = df1.loc[ (df1['extras']&1)!=0]["prev_satu"].sum()/1e+6   # in sec...
        #print(f"D... SATURATIONS {satu_n_ch} total: {Tsatu_ch} sec. at maximum")
        #print(f"D... SATURATIONS {satu_n_ch} total: {100*Tsatu_ch/interval:.2f} % at maximum")
        print(f"D... number of saturations: ",len( df1.loc[ (df1['extras']&1)==1] ) , f" {Tsatu_ch:.2f} sec. at max;  {100*Tsatu_ch/interval:.2f} % at max"   )

    print(f"D... Emin Emax       : {df1['E'].min()} ...  {df1['E'].max()}    /2^15={2**15}/")
    print(f"D... Tmin Tmax       : {tmin:.2f} ... {tmax:.2f} ; total time {interval:.2f} sec")
    print("_"*50,"BASICS END     CHANNEL ",chan)


    #print(df1)
    ###print("  range of neighbours in [us]:",dtusmin, dtusmax)

    if 'extras' in  df1:
        satu_n_ch = len(df1.loc[ (df1['extras']&1)!=0] )
        Tsatu_ch = df1.loc[ (df1['extras']&1)!=0]["prev_satu"].sum()/1e+6   # in sec...
        # print(f"D... SATURATIONS {satu_n_ch} total: {Tsatu_ch} sec. at maximum")
        #print(f"D... SATURATIONS {satu_n_ch} total: {100*Tsatu_ch/interval:.2f} % at maximum")
    else:
        satu_n_ch = None
        Tsatu_ch = None

    #len_zeroes,len_dzeroes,len_szeroes,len_izeroes,len_stazeroes,zoutput = ascgeneric.pd_detect_zeroes(df1, chan, TIMEWIN_US=4.48) # df1[ (df1.E==0) ]
    #print(zoutput)

    return df1















#------------------------------------------------------------------------

def enhance_by_dtus_and_next_E(df):
    #print("D... adding dtus and next_E columns")
    #print( freemem() )


    #print(f"D... broadening table with dt :next event  in t+dtus")
    df['dtus'] = (df.time.shift(-1)-df.time)*1000*1000
    df['dtus'] = df['dtus'].astype('float32') #
    dtusmin = df['dtus'].min()
    dtusmax = df['dtus'].max()
    df.fillna(99999, inplace =True)
    #print(f"D... range of time-differences (erlang) values min/max : {dtusmin:.3f} us ...  {dtusmax:.3f}  usec." )


    #print(f"D... broadening table with dt :prev event  = t - dtus (to enable search for standalones)")
    df['dtuspr'] = (df.time - df.time.shift(+1))*1000*1000
    df['dtuspr'] = df['dtuspr'].astype('float32') #
    dtusminpr = df['dtuspr'].min()
    dtusmaxpr = df['dtuspr'].max()
    df.fillna(99999, inplace =True)
    #print(f"D... range of time differences (erlang) values : {dtusminpr:.3f} us ...  {dtusmaxpr:.3f}  usec. PREVIOUS" )

    # df['dtus3'] = (df.time.shift(-2)-df.time)*1000*1000 # for triple
    # #df.fillna(9999, inplace =True)
    # df['dtus3'] = df['dtus3'].astype('float32') #
    # dtusmin3 = df['dtus3'].min()
    # dtusmax3 = df['dtus3'].max()
    # df.fillna(99999, inplace =True)
    # print(f"D... range of time differences (erlang) values : {dtusmin3:.3f} us ...  {dtusmax3:.3f}  usec. TRIPLE" )

    # df['dtus4'] = (df.time.shift(-3)-df.time)*1000*1000 # for triple
    # #df.fillna(9999, inplace =True)
    # df['dtus4'] = df['dtus4'].astype('float32') #
    # dtusmin4 = df['dtus4'].min()
    # dtusmax4 = df['dtus4'].max()
    # df.fillna(99999, inplace =True)
    # print(f"D... range of time differences (erlang) values : {dtusmin4:.3f} us ...  {dtusmax4:.3f}  usec. QUADRUPLE" )

    # df['dtus5'] = (df.time.shift(-4)-df.time)*1000*1000 # for triple
    # #df.fillna(9999, inplace =True)
    # df['dtus5'] = df['dtus5'].astype('float32') #
    # dtusmin5 = df['dtus5'].min()
    # dtusmax5 = df['dtus5'].max()
    # df.fillna(99999, inplace =True)
    # print(f"D... range of time differences (erlang) values : {dtusmin5:.3f} us ...  {dtusmax5:.3f}  usec. PENTUPLE" )




    #print(f"D... broadening table - next_E")
    df['next_E'] = (df.E.shift(-1))
    df.fillna(0, inplace=True)
    df['next_E'] = df['next_E'].astype('int32') #

    # print(f"D... broadening table - next_E3")  # for triple
    # df['next_E3'] = (df.E.shift(-2))
    # df.fillna(0, inplace=True)
    # df['next_E3'] = df['next_E3'].astype('int32') #

    # print(f"D... broadening table - next_E4")  # for triple
    # df['next_E4'] = (df.E.shift(-3))
    # df.fillna(0, inplace=True)
    # df['next_E4'] = df['next_E4'].astype('int32') #

    # print(f"D... broadening table - next_E5")  # for triple
    # df['next_E5'] = (df.E.shift(-4))
    # df.fillna(0, inplace=True)
    # df['next_E5'] = df['next_E5'].astype('int32') #




    #print(f"D... broadening table - next_ch")
    df['next_ch'] = (df.ch.shift(-1))
    df.fillna(0, inplace=True)
    df['next_ch'] = df['next_ch'].astype('int32') #

    if 'extras' in df.columns:
        #print(f"D... broadening table - prev_dtus SATU  previous time")
        df['prev_satu'] = (df.time - df.time.shift(+1))*1000*1000
        #  this says - if no saturation flag=> set 0
        df['prev_satu'] = np.where((df['extras']&1)==1, df['prev_satu'] , 0)
        #  same as
        # df.loc[  (df.extras&1)!=1, 'prev_satu' ] =  0
        #
        #
        df.fillna(0, inplace = True)
        df['prev_satu'] = df['prev_satu'].astype('float32') #

    #----------nice debug print ---------------------
    #print(df)
    #endblock()
    return df, dtusmin, dtusmax







#----------------------------------------------------------------------
def asc_stat_print( df, TIMEWIN_US = 4.2 , filename = "", TRGHOLD = 0):
#pd_detect_zeroes(df, channel,  TIMEWIN_US = 4.2 ):
    """
   2-dtus
    get number of zeroes, single zeroes, double zeroes, standalone zeroes
    -
      2nd phase - SZ and DZ - make one event from these PILEUPs
      3rd phase - create

    #len_zeroes,len_dzeroes,len_szeroes,len_izeroes,len_stazeroes,zoutput = ascgeneric.pd_detect_zeroes(df1, chan, TIMEWIN_US=4.48) # df1[ (df1.E==0) ]
    #print(zoutput)

    """
    print( f"D... detection of zeroes: WINDOW == {bg.red}{fg.white} {TIMEWIN_US:.1f} us {bg.default}{fg.default}")



    # vycisteny kanal, ----------------------------------- DFZ
    # I PREPARE SEVERAL VIEWS : only this channel :
    dfz = df[:-1] # WHY????????????
    chan_available = df['ch'].unique()
    if len(chan_available)>1:
        print(f"X... {fg.red} select ONE channel only - this is statistics over all channels{fg.default}")
        stat_asc(dfz)
        return None
    channel = chan_available[0]
    #index_names = dfz[  (dfz.ch!=channel)].index
    #dfz = dfz.drop(  index_names )
    zeroes = len(  dfz.loc[ (dfz.E==0) ] )
    #print( f" ... TOTAL ZEROES  = {zeroes}")

    # same as dfz for the moment -------------------------- DFALONE IS ALONE !!!!!!!!!!
    dfalone = df # drop the close (cluster) events to count standalones
    index_names = dfalone[ (dfalone.ch!=channel)].index
    dfalone = dfalone.drop( index_names )
    # STANDALONE EVENTS  - drop all events close to another event
    index_names = dfalone[ (dfalone.dtus<=TIMEWIN_US) | (dfalone.dtuspr<=TIMEWIN_US)  ].index
    dfalone = dfalone.drop( index_names) # NOT inplace, else it changes the df!!! # LAST REDEFINITION ALONE

    #print(f"i... purely isolated events {len(dfalone)}" )
    isoevents = len( dfalone )
    isozeroes = len( dfalone.loc[ (dfalone.E==0)  ])
    isononzeroes = len( dfalone.loc[ (dfalone.E!=0)  ])
    # print(f" ... purely isolated zeroes {stazeroes}")


    # playing on clusters---------------------------- DFZ IS CLUSTERS !!!!!!!!!!!
    # ----------- COMPLEMENT TO STANDALONES ... drop all isolated, keep clusters
    dfcluster = dfz[~dfz.isin(dfalone) ].dropna()
    # COMBINED ZEROES ... PAIRS treatment --- INSIDE dfcluster --- i look forward and check that nextE is not out of cluster
    # ilogic zeroes
    izeroes = len( dfcluster.loc[ ((dfcluster.E==0) & (dfcluster.next_E!=0))  & (dfz.dtus<=TIMEWIN_US)  ] )
    #print( f" ... ilogiczeroes = {izeroes}" )
    szeroes = len(dfcluster.loc[ ((dfcluster.E!=0) & (dfcluster.next_E==0)) & (dfz.dtus<=TIMEWIN_US)  ] )
    #print( f" ... singlezeroes = {szeroes} " )
    dzeroes = len(dfcluster.loc[ ((dfcluster.E==0) & (dfcluster.next_E==0))  & (dfz.dtus<=TIMEWIN_US) ] )
    #print( f" ... doblezeroes = {dzeroes} " )
    dnonzeroes = len(dfcluster.loc[ ((dfcluster.E!=0) & (dfcluster.next_E!=0))  & (dfz.dtus<=TIMEWIN_US) ] )

    nclusters =  len(dfcluster.loc[  (dfz.dtus>TIMEWIN_US) ] )


    #dfz.reset_index(inplace=True)
    # print(f" ... clustered events    = {len(dfz)}")

    #index_names = dfz[ (dfz.E!=0) ].index # drop nonzeroes
    #dfz = dfz.drop( index_names )
    #print(f" ... clustered where E=0 = {len(dfz)}")

    t1 = df['time'].max()
    t0 = df['time'].min()
    rate = len(df)/(t1-t0)
    # ---------------
    wind_erlang = erlang.erlang_cum(float(TIMEWIN_US)*1e-6, rate, 1)*100  # in %    bellow window 4.48us
    trgh_erlang = erlang.erlang_cum(float(TRGHOLD)*1e-6, rate, 1)*100  # in %    bellow window 4.48us

    satun_n_ch = None # n saturations
    Tsatu_ch = None # time max in satu
    Tsatu_pr = None # percent max in satu
    if 'extras' in df:
        #print(f"D... extras    present: {df1['extras'].unique()}... 1..satur,3..roll,4..reset,8..fake" )
        satu_n_ch = len(df.loc[ (df['extras']&1)!=0] )
        Tsatu_ch = df.loc[ (df['extras']&1)!=0]["prev_satu"].sum()/1e+6   # in sec...
        Tsatu_pr = 100*Tsatu_ch/(t1-t0)
        #print(f"D... SATURATIONS {satu_n_ch} total: {Tsatu_ch} sec. at maximum")
        #print(f"D... SATURATIONS {satu_n_ch} total: {100*Tsatu_ch/interval:.2f} % at maximum")

        #print(f"D... number of saturations: ",len( df.loc[ (df['extras']&1)==1] ) , f" {Tsatu_ch:.2f} sec. at max;  {Tsatu_pr:.2f} % at max"   )


    output = f"""________________________________________________________________________
 {filename:40s}        {t0:.2f} ... {t1:.2f} sec.
________________________________________________________________________
    D... total    events (chan={channel}) = {len(df):8d}
    D... isolated events (chan={channel}) = {isoevents:8d}
    D... clusterd events (chan={channel}) = {len(df)-isoevents:8d}  (in {nclusters} clusters)

    D... total    nonzero(chan={channel}) = {len(df)-zeroes:8d}
    D... isolated nonzero(chan={channel}) = {isoevents-isozeroes:8d}
    D... clusterd nonzero(chan={channel}) = {len(df)-zeroes-isononzeroes:8d}

    D... total    zeroes (chan={channel}) = {zeroes:8d} ~ {zeroes/len(df)*100:5.2f}% {fg.green}/DT-caen/{fg.default}
    D... isolated zeroes (chan={channel}) = {isozeroes:8d} ~ {isozeroes/len(df)*100:5.2f}%  [glitches]
    D... clusterd zeroes (chan={channel}) = {zeroes-isozeroes:8d}
    D... single   zeroes (chan={channel}) = {szeroes:8d} (E-0)  /{100*szeroes/len(df):.2f} %/
    D... double   zeroes (chan={channel}) = {dzeroes:8d} (0-0)  /{2*100*dzeroes/len(df):.2f} %/ {fg.yellow}/?clusters/{fg.default}
    D... ilogic   zeroes (chan={channel}) = {izeroes:8d} (0-E)  /should be 0/
    D... double   E      (chan={channel}) = {dnonzeroes:8d} (E-E)  /should be 0/

    D... blind erlang    (chan={channel}) = {wind_erlang:.2f} %  {fg.green}/DT from ERLANG/{fg.default} {TIMEWIN_US:.1f} us
    D... blind erlang    (chan={channel}) = {trgh_erlang:.2f} %  {fg.green}/DT from ERLANG/{fg.default} {TRGHOLD:.1f} us

    D... n_sat,tmax_sat  (chan={channel}) = {satu_n_ch:8d}, {Tsatu_ch:.2f} s. {Tsatu_pr:.1f} %  {fg.green}/DT from ERLANG/{fg.default} {TRGHOLD:.1f} us
all blind10==lost ~~ all zeroes+blind0.5==lost; maybe some 0 are exagerated @ortec; NEED [satur-median] estimation HERE
_______________________________________________________________________________
"""
    # {fg.red}D... double zeroes (chan={channel}) = {dzeroes:8d} ~  {2*dzeroes/len(df)*100:5.2f}% ( % counted correctly 2x){fg.default}
    # D... => zrs@HigherClusters  = {zeroes-isozeroes-2*dzeroes:8d}  (zeroes in longer than 2 clusters)
    # D... ______________________ end of zero detection
    # D...  window {TIMEWIN_US} us   ....  crutial for correct double zeroes detection

    print(output)
    #     # returns length for now
    #     # endblock()
    return zeroes/len(df)*100
    return None # zeroes,dzeroes,szeroes,izeroes,isozeroes,output




def cut_asc(df, t_from, t_to):
    t0 = float(t_from)
    t1 = float(t_to)
    df1 = df.loc[ (df.time>t0)&(df.time<t1)]
    return df1



#------------------------------------------------------ FROM GENERAL READVMEOFFLINE
def fill_hist( his, narr):
    """
    FILL TH1D(?) histogram by  np.histogram data.
    simple way bin by bin
    """
    #print(f"D... filling TH  bin by bin ({narr.shape[0]}) bins")
    if len(narr.shape) == 1: # 1D
        for i in range(narr.shape[0]):
            #if narr.shape[0]<6:
            #    print(i, narr[i], "  ", end=";")
            # we start with 0, bin starts with 1 in ROOT
            his.SetBinContent( i+1, narr[i] )
        his.SetEntries( narr.sum() ) # complete with setting entries == integral

    elif narr.shape[1] == 2: #  2D not tested
        print("X... 2D not tested ... returning None", narr.shape )
        return None
        for i in range(narr.shape[0]):
            for j in range(narr.shape[1]):
                his.SetBinContent( i+1, j+1, narr[i][j] )
    #print("D... filling TH DONE")

#------------------------------------------------------------------------  FROM GENERAL READVMEOFFLINE
def column_to_newhisto( dfcol , binmax = 32*1024, himax=32*1024, savename = None, hname = "h1", writeondisk = False , writetxt = False, calibration = None ):
    """
    df["E"] ... creates NEW histogram from df column
    - converts to numpy array; creates HIST;
    - saves txt
    - updates allhist.root too
    """
    narr = dfcol.to_numpy()
    print(f"D... ========= creating histo {hname} ======= len={len(narr)}; bins=", binmax)
    #print("D... len=", len(narr), narr.dtype, "...  ndarray to np histo")
    #print(dfcol)
    #print(narr)
    his,bins = np.histogram(narr,bins=binmax,range=(0.0,himax) )
    #    print("D... histo:", his)
    #    print("D... bins :", bins)
    del narr

    #
    basename = hname
    hname2 = hname
    if hname2.find(";"):
        hname2 = hname2.split(";")[0]
    if savename is not None:
        basename = os.path.basename(savename) # title without path...

    #basename = f"{basename}; [us]"
    print( f"th1f = ROOT.TH1D( /{hname2}/, /{basename}/, /{binmax}/, 0 , /{himax}/ ) ")
    th1f = ROOT.TH1D(hname2, basename, binmax, 0.0 , himax )
    #for i in range(len(his)):
    #    th1f.SetBinContent( i, his[i] )
    #print("D... narr=",narr)

    #print(narr.shape  ,len(narr.shape) )
    #print(narr.shape[0] )
    fill_hist( th1f,  his )

    if not calibration is None:
        CALA = calibration[0]
        CALB = calibration[1]
        Cxmin = th1f.GetXaxis().GetXmin() * CALA + CALB
        Cxmax =th1f.GetXaxis().GetXmax() * CALA + CALB
        th1f.GetXaxis().SetLimits(Cxmin,Cxmax)
    #th1f.Print()

    if (savename is not None) and  writeondisk:
        if writetxt:
            print(f"D...          text spectrum = {savename}")
            np.savetxt(savename, his, fmt="%d")
        # rootname = os.path.splitext(savename)[0]+".root"
        rootname = os.path.dirname(savename)
        if len(rootname)>3:
            rootname+="/"
        rootname+="all_histograms.root"
        print("D...          root spectum at", rootname, "name=", hname )
        f = ROOT.TFile(rootname,"UPDATE")
        # print(f"D...          {th1f.GetName():13s} @ {rootname}")
        th1f.Write()
        f.Close()
    #endblock(nomem=True)
    return th1f # WAS RETURNIG HIS -- NUMPY ARRAY



def histo_erlang( df1 , hname = 'erlang2' ):
    hise2 = column_to_newhisto(df1['dtus'].loc[(df1.E!=0)],
    #hise2 = column_to_newhisto(df1['dtus'],
                               binmax=200, himax=20,
                                     hname = f"{hname}; dt [us]")
    hise2.Print()
    return hise2

def histo_dzeroes( df1 , hname = 'dzeroes' ):
    hisdz = column_to_newhisto(df1['dtus'].loc[(df1.E==0)&(df1.next_E==0)],
                                       binmax=200, himax=20,
                                  hname = f"{hname}; dt [us]")

    hisdz.Print()
    return hisdz

def histo_szeroes( df1 , hname = 'szeroes' ):
    hissz = column_to_newhisto(df1['dtus'].loc[(df1.E!=0)&(df1.next_E==0)],
                                       binmax=200, himax=20,
                                  hname = f"{hname}; dt [us]")

    hissz.Print()
    return hissz

def histo_energy( df1 , hname = 'energy' , calibration = None ):
    units = "[ch]"
    maxc = df1.E.max() # should not be more than int 2**15

    if maxc > 2**15-1:
        maxc = 2**14
    maxc = 2**15
    print("D... max E", maxc)

    #chan0 = df1.ch.min()
    chan = df1.ch.max()

    if calibration is not None:
        units = "[keV]"
        print("D... calibration",calibration)
        if calibration[0]==1 and calibration[1]==0:
            units  = "(chan)"
    else:
        units = "[chan]"
    hisene = column_to_newhisto(df1['E'].loc[(df1.E!=0)],
                                binmax=maxc, himax=maxc,
                                hname = f"{hname}; Energy {units}; channel {chan}", calibration = calibration)
    hisene.Print()
    return hisene


def histo_time( df1 , hname = 'time' , lowe=-99999, hie=99999 ):
    maxc = df1.time.max()
    c = df1.ch.max()
    hist = column_to_newhisto(  df1['time'].loc[ (df1.E!=0) & (df1.E>=lowe) & (df1.E<=hie) ],
                                       binmax=int(maxc)+1, himax=maxc,
                                  hname = f"{hname}; time [s]; chan {c}")


    #print( df1.loc[ (df1.E!=0) & (df1.E>=lowe) & (df1.E<=hie) ] )
    hist.Print()
    return hist



#!/usr/bin/env python3

from fire import Fire
import ROOT
import urllib.request
import json
import time
import datetime as dt
import numpy as np
import os
from console import fg,bg,fx
import socket
import threading
import re # decode title

class Histogram:
    """
    start2collect ... it loops.  Kills with do_run=False

    """

    def set_step(self, step):
        self.STEP = step
        print(f"i... poll step set to /{step}/")


    def __init__(self, url, spectrum="b0c00", peaksfile=None, calibrationfile=None, bgfile=None ):
        """
        init & first draw; calls self.wait at the end == runs in loop
        """

        self.started = None #  dt.datetime.now()
        self.tcp_fetch_time = None
        self.measurement = "hpge" # string in INFLUXDB

        self.realtime = None

        self.url = f'http://{url}:9009/{spectrum}/root.json'
        self.spectrumname = spectrum
        print(f"i... url = {self.url}")

        self.size = 2**15 # histogram size
        self.xmin = 0 # no calibration at the start, avoid crash in get_calib
        self.xmax = self.size

        self.maximum = 10 # default of histomax, but overriden immediat
        self.N_EMIN = 20 # 20keV noise end by definition
        self.e_window = 15 # +- Window to ZOOM and Analyze


        self.first_influx_omit = True # ONETIME TRIG

        self.bgfile = bgfile # others are used later....  BACKGROUND FILE
        self.bgfile = "bg88320.txt" # others are used later....  BACKGROUND FILE
        self.bgtime = 88320
        #-------------------- timing of display

        self.STEP = 2 # seconds THIS IS COLLECTION INTERVAL


        self.SECONDS = 10
        self.MINUTE = 60 # 5minutes (300)
        #self.MINUTE = 60 #

        self.HOUR = 3600
        #self.HOUR = 60


        # calibration that comes from GetXaxis()
        #self.xmin = 0
        #self.xmax = self.size

        #self.store_limit = 3600 # seconds
        self.store_limit = 3*3600 # seconds

        self.first_fetch = True # avoid 1st fetch diff
        #------------------------------------------------ init NumPy array
        self.h1np = np.zeros( self.size+2)
        self.h1np_prev = np.zeros( self.size+2)
        self.h1np_diff = np.zeros( self.size+2)

        #---------------------- store this got get()
        self.fflag = False # fetch flag
        now1 = dt.datetime.now() # fetch flag
        time.sleep(0.2)
        self.runtime = dt.datetime.now() - now1 # fetch flag
        # print("D... init end")
    #------------------------------------------------------------

    #i keep all the history histos here in this object
    def get_lenhistos(self):
        return len(self.histos)

    def get_limits_calib(self):
        return self.xmin,self.xmax

    def get_limits_bins(self):
        return 0,self.size


    def get_partial_time_histo(self, age):
        ele=None
        dist=999999
        #print(f"D... demanded age = {age}")
        for k in self.histos.keys():
            if ele is None: ele = k
            agedif = abs(self.histos[k][0] - age)
            #print(f"D... {k} - histo = {self.histos[k]}")
            if agedif < dist:
                ele = k
                dist = agedif
        #print(f"D... retuning closest distance  == {dist}")
        return self.histos[ele][1], dist


    # earlier set_runtime...
    def set_started_before(self, seconds, silent = False):
        if not silent:  print("i... started was=",self.started,"  calulating from ",seconds)
        self.started = dt.datetime.now() - dt.timedelta(seconds=seconds)
        # there was some worry aobut resistance to clear....
        # ... spectrum jumps...
        # self.runtime = dt.timedelta(seconds=seconds)
        if not silent:  print("i... newly       ",self.started, "runtime=",self.runtime )
        #self.runtime =


    def get_started(self):
        return self.started


    def get_runtime(self):
        #print( "D... get runtime:", self.runtime, type(self.runtime)  )
        # print(f"D... realtime in get_runtime  = {self.realtime}")
        if type(self.runtime) is dt.timedelta:
            if self.runtime.total_seconds()<=0:
                self.runtime = dt.timedelta(seconds=1)

        if self.realtime is not None: # TITLE is used .....
            #print( "D... realtime=", self.realtime  ,   self.runtime )
            if abs(self.realtime  -self.runtime.total_seconds() )>15: # DEIFFERENCE 15 seconds ........
                print(f"{fg.red}START TIME RESET !!!!!!!!!!{fg.default}")
                self.set_started_before( self.realtime ) # RESET RUNTIME, SINCE THERE IS A 15s difference
            return dt.timedelta( seconds=self.realtime ) # taken from TITLE
        else:
            return self.runtime
        return 1


    def get_h1np(self):
        return self.h1np

    def get_h1np_diff(self):
        return self.h1np_diff

    def get_deltat(self):
        return self.deltat


    def fetch_flag(self):
        """
        Let know ONLY ONCE that there was a fetch
        """
        if self.fflag:
            self.fflag = False
            return True
        return False



    def fetch_histo(self):
        """
        Every fetch, numpy arrays are filled /  TH1F content is set too
        h1np and h1np_prev and h1np_diff    are kept
        """


        #print(f"i... fetching ...{self.url}", end="\n")

        fstart = dt.datetime.now()
        ok = False
        try:
            with urllib.request.urlopen(self.url, timeout=5) as response:
                jhis = response.read().decode("utf8")
            ok = True
            #print("i... ok")
        except:
            ok = False
        self.tcp_fetch_time = (dt.datetime.now() - fstart).total_seconds()


        if not ok :
            print(f"!... no connection {self.url}")
            return False

        jhis = json.loads(jhis)
        # dict_keys(['_typename', 'fUniqueID', 'fBits', 'fName', 'fTitle', 'fLineColor', 'fLineStyle', 'fLineWidth', 'fFillColor', 'fFillStyle', 'fMarkerColor', 'fMarkerStyle', 'fMarkerSize', 'fNcells', 'fXaxis', 'fYaxis', 'fZaxis', 'fBarOffset', 'fBarWidth', 'fEntries', 'fTsumw', 'fTsumw2', 'fTsumwx', 'fTsumwx2', 'fMaximum', 'fMinimum', 'fNormFactor', 'fContour', 'fSumw2', 'fOption', 'fFunctions', 'fBufferSize', 'fBuffer', 'fBinStatErrOpt', 'fStatOverflows', 'fArray'])


        #print(jhis.keys() )

        # print( jhis["fTitle"])

        grp = re.search(r'rt=([\d\.]+)$',jhis["fTitle"])  # rt=3600 at the END of title ...
        # print(grp)
        if grp is not None and len(grp.groups())>0:
            self.realtime = float(grp.groups()[0].split("=")[-1])
            #print( f" realtime from title grp ={self.realtime}" )

        # THIS STEERS CALIBRATION ..... limits_calib
        self.xmin = jhis['fXaxis']['fXmin']
        self.xmax = jhis['fXaxis']['fXmax']

        #print( jhis.keys() )
        helem = jhis['fArray']

        # print("D... helem in fetch: len,0,1 ...", len(helem) , helem[0], helem[-1] )

        self.h1np_prev = 1.0*self.h1np # save  previous
        self.h1np = np.array( helem )

        #---------protection in case of  reset
        if self.h1np.sum()< self.h1np_prev.sum():
            print(f"i... {fg.red}RESET? I CLEAN THE PREV. HISTOGRAM{fg.default}")
            self.h1np_prev = self.h1np_prev*0
            self.histos = {} # !!!
            # BUG?
            self.set_started_before( 1 ) # BUG?

        if self.first_fetch:
            self.first_fetch = False
            self.h1np_diff = self.h1np*0
        else:
            self.h1np_diff = self.h1np - self.h1np_prev

            # print("D... diff after substract", self.h1np_diff)
            # print("D... h1np and prev@ substract",  self.h1np, self.h1np_prev)
            # underflow should be retained

        #sum = 0

        #self.maximum = max(helem)

        #print(f"D... end fetch {sumnow-sumprev}  {sumnow}, {sumprev}")
        return True




    def start2collect(self):
        """
        We reset histos DICT and start to collect all
        dict members = list [ deltat, histo ]
        ON RESTART - it resets the self.started
        """
        self.histos = {}

        # i think these are starting values---------------------

        deltas = np.zeros( 2 ) +0.2
        avgdeltat = 0.2
        avgoverhead = 0.02
        overheads = np.zeros(2) + 0.02
        lastnow = dt.datetime.now() # for the 1st

        #for i in range(20):
        self.thrmyself = threading.current_thread()

        #while True: THREAD WaTCHES the FLAG from INSIDE... it can be stopped from outside
        while getattr( self.thrmyself, "do_run", True ):
            # print( "getattr FETCH",  getattr(self.thrmyself, "do_run", True) )

            start = dt.datetime.now()
            self.fetch_histo()
            now = dt.datetime.now() # NOW = AFTER HISTO IS DELIVERED

            if (now-lastnow).total_seconds() > self.STEP:
                overheads = np.append( overheads,   (now-lastnow).total_seconds() - self.STEP )

            deltat = (now - start).total_seconds()
            deltas = np.append( deltas, deltat )

            avgdeltat = deltas.mean()

            sumnow = self.h1np.sum()
            sumprev = self.h1np_prev.sum()
            # -------- if no change or smaller SUM (clashes with fetch reset)
            #if (sumnow<=sumprev) or (sumnow==0) or (sumprev==0):
            if (sumnow<=sumprev) :
                # self.started=now # this doesnt reset really
                self.set_started_before( 1 ) # BUG?
                print("I... START TIME RESET")
            # -------- if no change or smaller SUM (clashes with fetch reset)
            if (sumnow<=sumprev) or (sumnow==0):
                self.histos = {}

            # ---------runtime, that is resistant to clear or no count
            #  HERE I PREPARE REALTIME AND RUNTIME (if no title info available)
            #   - this (together wth get_) can keep self.started  GOOD - TITLEvar works...

            if self.started is None:
                if self.tcp_fetch_time is not None:
                    if self.realtime is not None: # set from TITLE rt=
                        # print(f"D... {fg.yellow}1st tuning of starttime from title realtime{fg.default}")
                        self.set_started_before( self.realtime + self.tcp_fetch_time/2)
                    else:
                        # print(f"D... {fg.yellow}1st tuning of starttime from runtime{fg.default}")
                        self.set_started_before( self.runtime.total_seconds() + self.tcp_fetch_time/2 ) # guess

            if self.realtime is None: # do it from our time
                if self.started is not None:
                    self.runtime = now - self.started
            else: # DO IT FROM TITLE -
                self.runtime = dt.timedelta(seconds = self.realtime)
                # -nothing,,
            # if self.runtime < 0:
            #     if self.realtime is not None:
            #         self.runtime = now - self.started

            #print(self.started)
            #print(self.runtime)



            # print(f"i... {str(self.runtime)[:-4]} dt= {deltat:.2f} [avg={avgdeltat:.2f}], ovh~{overheads.mean():.2f} #={len(self.histos)}")

            self.histos[ now ] = [ 0, self.h1np ]



            # least_dist = 1000*1000
            # least_who = now
            # --------- set proper delta times
            removekeys = []
            for k,v in self.histos.items():
                age = (now - k).total_seconds()
                if age>self.store_limit:
                    #self.histos.pop(k)
                    removekeys.append(k)
                else:
                    self.histos[k][0] = age
            for key in removekeys:
                self.histos.pop(key)



            self.deltat = ( now - lastnow ).total_seconds()
            lastnow = now



            # ALL PREPARED.......... announce and sleep.......
            self.fflag = True


            # SLEEP WISELY----------------------------------------------------
            #--------------------------calculate deltas(web) and overhead(code)
            # and sleep.
            if avgdeltat/2<self.STEP:
                slp = self.STEP - avgdeltat/2 - overheads.mean()
                if slp<0.1: slp = 0.1
                time.sleep( slp )
            if len(overheads)>20:
                overheads = np.delete( overheads,0)
            if len(deltas)>20:
                deltas = np.delete( deltas,0) # delete 1st member






if __name__=="__main__":
    b0c01 = Histogram("192.168.250.60", "b0c01")
    b0c01.start2collect()
    print("__________________-")



#!/usr/bin/env python3
"""
Container for Histograms - easier to handle than gDirectory...
"""
from fire import Fire

histograms = {} #  ----- histograms created by system
canvases = {}

range_and_peaks = {}  # for id. and fit
  # range_and_peaks["spe"] = "" #spectrum name
  # range_and_peaks["pks"] = "" #list np

gregory_started = None # launch time

# calibration via "c" comand; use global on change and import
ECALIB = []
ECALIBa = 1
ECALIBb = 0

ECALIBdb = {}
ECALIBdb['40K'] = 1460.82
ECALIBdb['214BIB'] = 609.31
ECALIBdb['214BIA'] = 1764.49
ECALIBdb['208TL'] = 2614.51
ECALIBdb['60COA'] = 1173.2
ECALIBdb['60COB'] = 1332.5
# ECALIBdb[''] =

# last fit result - useful bin for calib....
last_fit_res = {}

is_zoomed = [ False, 0.0 ] # yes or no; middle_energy

def main():
    print()

if __name__=="__main__":
    Fire(main)



#!/usr/bin/env python3
import os
from setuptools import setup, find_packages

#-----------problematic------
def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()

import os.path

def readver(rel_path):
    here = os.path.abspath(os.path.dirname(__file__))
    with open(os.path.join(here, rel_path), 'r') as fp:
        return fp.read()

def get_version(rel_path):
    for line in readver(rel_path).splitlines():
        if line.startswith('__version__'):
            delim = '"' if '"' in line else "'"
            return line.split(delim)[1]
    else:
        raise RuntimeError("Unable to find version string.")

setup(
    name="gregory_online",
    description="Online watching HTTP server of ROOT on port 9009",
    author="me",
    author_email="jaromrax@gmail.com",
    license="GPL2",
    version=get_version("gregory_online/version.py"),
    packages=['gregory_online'],
    package_data={'gregory_online': ['data/*']},
    long_description=read('README.md'),
    long_description_content_type='text/markdown',
    scripts = ['bin/gregory_online'],
    install_requires = ['fire','scipy','pyfromroot','pytermgui','readchar','fastnumbers','sshkeyboard','terminaltables','fastparquet','tables','simple_term_menu',"tdb_io"],
)
